/*
 * Socket server for JVisu
 *
 * Copyright (C) 2003 Thomas Gleixner <tglx@linutronix.de>
 *		
 * PURPOSE
 * 
 * This is a lightweight, single threaded multiconnection server for
 * data visualization based on JVisu applets. 
 * 
 * It's single threaded as it is intended to run on as much platforms
 * as possible. For uClinux a single threaded server is the best choice.
 * On standard linux platforms this concept does not hurt.
 * 
 * All you have to provide is a structure with functions for symbol 
 * resolving and variable write. Variable write can be NULL, get_symbol 
 * must be a valid function.
 * 
 * The server is running as long as the global variable jvisu_stop is zero.
 * You can stop the server by modifying jvisu_stop.
 * Signal handling is not done inside this code to make it as portable as
 * possible. You can do this in the thread which calls this code.
 *
 * Note that most communication related data is wrapped with special 
 * datatypes which let you not access this data directly. They ensure
 * endianess conversion. Use the appropriate conversion macro.
 */

/*
 * LICENCING INFORMATION
 *
 * JVisu is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the 
 * Free Software Foundation
 *
 * JVisu is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with JVisu; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * EXCEPTION 
 * for JVisu server files (jvisu.c and jvisu.h) only, which are referenced
 * by the term 'these files' in the following section.
 * As a special exception, if other files instantiate templates or use
 * macros or inline functions from these files, or you compile these
 * files and link them with other works to produce a work based on these
 * files, these files do not by themselves cause the resulting work to be
 * covered by the GNU General Public License. However the source code for
 * these files must still be made available in accordance with section (3)
 * of the GNU General Public License.  This exception DOES NOT invalidate 
 * any other reasons why a work based on these files might be covered by 
 * the GNU General Public License.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include <time.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>

#include "jvisu.h"

/* debug */
#define D1(x) if (jvisu_debug) x

/* stop */
static int	jvisu_stop;

/* debug */
static int 	jvisu_debug;

/* user supplied functions */
static struct jvisu_ops	*jops;

/* list of active sockets */
static struct jvisu_sock 	actsd[MAX_CONNECTIONS + 1];
static int			connections;

/* file descriptor sets */
static fd_set			rfds, wfds;

/* registered variables */
static struct list_head		varlist;
static int			registeredvars;

/* timekeeping */
static struct timeval		tact;	

/* prototypes */
static void unregister_vars (struct jvisu_sock *sd);
static void shutdown_sock (struct jvisu_sock *sd);
static void check_variables (void);
static int start_socket (int port);
static int do_accept (int sd);
static int analyse_command (struct jvisu_sock *sd);
static int process_request (struct jvisu_sock *sd);
static void check_fdset (void);

static const char *prompt = "jvisu:";

/** 
 *	Unregister all variables, which are registered for a given 
 *	socket connection
 *  
 *	@param sd	socket description structure
 */
static void unregister_vars (struct jvisu_sock *sd)
{
	struct list_head *lact, *ltmp;
	
	/* remove from list, free memory */
	list_for_each_safe(lact, ltmp, &varlist) {
		struct jvisu_var *var = list_entry(lact, struct jvisu_var, list);
		if (var->sd->sd != sd->sd)
			continue;
		registeredvars--;
		list_del (&var->list);
		free (var);
	}
}


/** 
 * 	Shut down a socket and unregister variables
 * 	
 *	@param sd	socket description structure
 */
static void shutdown_sock (struct jvisu_sock *sd)
{
	D1 (printf ("%s shutdown socket: %d\n", prompt, sd->sd));

	unregister_vars (sd);
	connections--;
	FD_CLR (sd->sd, &rfds);
	FD_CLR (sd->sd, &wfds);
	close (sd->sd);
	sd->sd = -1;
}


/** 
 *	Retrieve a boolean variable from the application; do typecasting 
 *	from internal format
 *
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline u_char get_bool (struct jvisu_var *var)
{
	switch (var->inttype) {
		case VAR_TYPE_BOOL:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case VAR_TYPE_SHORT:
			return (*((uint16_t *) var->data)) ? 1 : 0;
		case VAR_TYPE_INT:
			return (*((uint32_t *) var->data)) ? 1 : 0;
		case VAR_TYPE_FLOAT:
			return (*((float *) var->data)) ? 1 : 0;
		case VAR_TYPE_DOUBLE:
			return (*((double *) var->data)) ? 1 : 0;
		case VAR_TYPE_STRING:	
		case VAR_TYPE_BYTE:
			return (*((u_char *) var->data)) ? 1 : 0;
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a byte variable from the application; do typecasting 
 *	from internal format
 *	
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline u_char get_byte (struct jvisu_var *var)
{
	switch (var->inttype) {
		case VAR_TYPE_BOOL:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case VAR_TYPE_SHORT:
			return (u_char) (*((uint16_t *) var->data));
		case VAR_TYPE_INT:
			return (u_char) (*((uint32_t *) var->data));
		case VAR_TYPE_FLOAT:
			return (u_char) (*((float *) var->data));
		case VAR_TYPE_DOUBLE:
			return (u_char) (*((double *) var->data));
		case VAR_TYPE_STRING:	
		case VAR_TYPE_BYTE:
			return (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a short variable from the application; do typecasting
 *	from internal format
 *	
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline short get_short (struct jvisu_var *var)
{
	switch (var->inttype) {
		case VAR_TYPE_BOOL:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case VAR_TYPE_SHORT:
			return (*((uint16_t *) var->data));
		case VAR_TYPE_INT:
			return (short) (*((uint32_t *) var->data));
		case VAR_TYPE_FLOAT:
			return (short) (*((float *) var->data));
		case VAR_TYPE_DOUBLE:
			return (short) (*((double *) var->data));
		case VAR_TYPE_STRING:	
			return (short) atoi ((u_char *) var->data);
		case VAR_TYPE_BYTE:
			return (short) (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a integer variable from the application; do typecasting 
 *	from internal format
 *	
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline int get_int (struct jvisu_var *var)
{
	switch (var->inttype) {
		case VAR_TYPE_BOOL:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case VAR_TYPE_SHORT:
			return (int) (*((uint16_t *) var->data));
		case VAR_TYPE_INT:
			return (*((uint32_t *) var->data));
		case VAR_TYPE_FLOAT:
			return (int) (*((float *) var->data));
		case VAR_TYPE_DOUBLE:
			return (int) (*((double *) var->data));
		case VAR_TYPE_STRING:	
			return atoi ((u_char *) var->data);
		case VAR_TYPE_BYTE:
			return (int) (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a float variable from the application; do typecasting 
 *	from internal format
 *
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline float get_float (struct jvisu_var *var)
{
	switch (var->inttype) {
		case VAR_TYPE_BOOL:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case VAR_TYPE_SHORT:
			return (float) (*((uint16_t *) var->data));
		case VAR_TYPE_INT:
			return (float) (*((uint32_t *) var->data));
		case VAR_TYPE_FLOAT:
			return (*((float *) var->data));
		case VAR_TYPE_DOUBLE:
			return (float) (*((double *) var->data));
		case VAR_TYPE_STRING:	
			return (float) atof ((u_char *) var->data);
		case VAR_TYPE_BYTE:
			return (float) (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a double variable from the application; do typecasting 
 *	from internal format
 *	
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline double get_double (struct jvisu_var *var)
{
	switch (var->inttype) {
		case VAR_TYPE_BOOL:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case VAR_TYPE_SHORT:
			return (double) (*((uint16_t *) var->data));
		case VAR_TYPE_INT:
			return (double) (*((uint32_t *) var->data));
		case VAR_TYPE_FLOAT:
			return (double) (*((float *) var->data));
		case VAR_TYPE_DOUBLE:
			return (*((double *) var->data));
		case VAR_TYPE_STRING:	
			return atof ((u_char *) var->data);
		case VAR_TYPE_BYTE:
			return (double) (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Check if variables have been added or changed
 */
static void check_variables (void)
{
	struct list_head 	*lact, *ltmp;
	struct timeval		tnow, tdiff;
	int			ms;
	
	/* read current time */
	gettimeofday (&tnow, NULL);
	tdiff.tv_sec = tnow.tv_sec - tact.tv_sec;
    	tdiff.tv_usec = tnow.tv_usec - tact.tv_usec;
	if (tdiff.tv_usec < 0) {
		tdiff.tv_sec--;
		tdiff.tv_usec += 1000000;
	}
	/* convert to ms */
	ms = (tdiff.tv_sec * 1000) + (tdiff.tv_usec / 1000);
	if (ms <= 0)
		return;

	/* save current time and take care of the usec remainder */
	tact = tnow;
	tact.tv_usec += tdiff.tv_usec % 1000;
	if (tact.tv_usec >= 1000000) {
		tact.tv_sec++;
		tact.tv_usec -= 1000000;
	}
	
	/* walk through the list */
	list_for_each_safe(lact, ltmp, &varlist) {
		size_t			len;
		int 			res, force;
		struct jvisu_sndvar	*snd;
		struct jvisu_var 	*var = list_entry(lact, struct jvisu_var, list);
		var->elapsed += ms;
		/* time elapsed ? */
		if (var->elapsed < var->rate)
			continue;
		/* transfer in progress ? */
		if (var->sd->status || (var->modifytype & VAR_WRITEONLY)) 
			continue;
			
		snd = (struct jvisu_sndvar *) var->sd->buf;
		snd->hdr.cmd = CMD_SEND_VAR;
		snd->hdr.type = (u_char) var->type;
		len = sizeof (struct jvisu_sndvar);
		snd->idx = cpu_to_je32 (var->idx);
		
		force = (var->status & VAR_STAT_INIT) ? 0 : 1;
		
		switch (var->type) {
		case VAR_TYPE_BYTE: {
			u_char act = get_byte (var);
			if ((var->cmp.b) == act && !force)
				continue;
			var->cmp.b = act;
			len++;
			snd->data.b[0] = act;
			break;
		}	
		case VAR_TYPE_SHORT: {
			short act = get_short (var);
			if ((var->cmp.s) == act && !force)
				continue;
			var->cmp.s = act;
			len += sizeof (short);
			snd->data.s[0] = cpu_to_je16 (act);
			break;
		}	
		case VAR_TYPE_INT: {
			int act = get_int (var);
			if ((var->cmp.i) == act && !force)
				continue;
			var->cmp.i = act;
			len += sizeof (int);
			snd->data.i[0] = cpu_to_je32 (act);
			break;
		}	
		case VAR_TYPE_FLOAT: {
			union { float f; uint32_t i; } act;
			act.f = get_float (var);
			if ((var->cmp.f) == act.f && !force)
				continue;
			var->cmp.f = act.f;
			len += sizeof (float);
			snd->data.i[0] = cpu_to_je32 (act.i);
			break;
		}	
		case VAR_TYPE_DOUBLE: {
			union { double d; uint64_t i; } act;
			act.d = get_double (var);
			if ((var->cmp.d) == act.d && !force)
				continue;
			var->cmp.d = act.d;
			len += sizeof (double);
			snd->data.d[0] = cpu_to_je64 (act.i);
			break;
		}	
		case VAR_TYPE_BOOL: {
			u_char act = get_bool (var);
			if ((var->cmp.b == act) && !force)
				continue;
			var->cmp.b = act;
			len++;
			snd->data.b[0] = act;
			break;
		}
		/* TYPE String provides no typecasting ATM
		 * As I have no real good idea to track changes on strings
		 * due to variable length and content I decided to transfer strings 
		 * in the given interval. 
		 * If you want to have change triggered string display 
		 * then use a index variable and provide the strings in a file
		 * which can be retrieved by the applet 
		 * If you are still not happy provide a better solution :)
		 */
		case VAR_TYPE_STRING:
			len += strlen ((char *)var->data);
			strcpy (snd->data.b, (char *)var->data);
			break;
		/* TYPE Stream is a special case. Check, if the length variable is
		 * > 0. Send out the bytes and set it to zero
		 */
		case VAR_TYPE_STREAM: {
			struct jvisu_varstream *p = (struct jvisu_varstream*) var->data;
			if (!p->len)
				continue;
			len += p->len;
			D1 (printf ("%s stream: Send %d bytes (total %d)\n", prompt, p->len, len));
			memcpy (snd->data.b, p->data, (size_t)p->len);
			p->len = 0;
			break;
		}
		case VAR_TYPE_BYTEARRAY: {
			int i;
			char *p = (char *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (p[i] != var->cmp.p[i]) {
					force = 1;
					break;
				}
			}		
			if (!force)
				continue;
			memcpy (snd->data.b, var->data, (size_t) var->arraylen);
			len += var->arraylen;
			break;
		}
		default:
			continue;
		}
		/* send value */
		snd->hdr.len = cpu_to_je16 ((ushort) len);
		res = write (var->sd->sd, var->sd->buf, len);
		if (res < 0) {
			if (res != EAGAIN) {
				shutdown_sock (var->sd);
				continue;
			}
			res = 0;
		}
		/* all sent ? */
		if (res == len) {
			var->elapsed = 0;	
			var->status |= VAR_STAT_INIT;
			continue;
		}
		fprintf (stderr, "%s FIXME: Send has to be delayed\n", prompt);	
		var->sd->bufpoi = res;
		var->sd->cmdlen = len;
		var->sd->status	= STAT_WRITE;
	}	
}


/** 
 *	Write a boolean variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_bool (struct jvisu_var *var, union jvisu_intvar *dat)
{
	int res;

	switch (var->type) {
		case VAR_TYPE_BOOL:
		case VAR_TYPE_STRING:	
		case VAR_TYPE_BYTE:
			res = dat->b[0] ? 1 : 0;
			break;
		case VAR_TYPE_SHORT:
			res = dat->s[0] ? 1 : 0;
			break;
		case VAR_TYPE_INT:
			res = dat->i[0] ? 1 : 0;
			break;
		case VAR_TYPE_FLOAT:
			res = (dat->f[0] != 0.0) ? 1 : 0;
			break;
		case VAR_TYPE_DOUBLE:
			res = (dat->d[0] != 0.0) ? 1 : 0;
			break;
		default:
			return;
	}
	if (res)
		*((u_char *) var->data) |= var->bitmask;
	else	
		*((u_char *) var->data) &= ~var->bitmask;
}


/** 
 *	Write a byte variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_byte (struct jvisu_var *var, union jvisu_intvar *dat)
{

	switch (var->type) {
		case VAR_TYPE_BOOL:
		case VAR_TYPE_BYTE:
			*((u_char *) var->data) = dat->b[0];
			break;
		case VAR_TYPE_SHORT:
			*((u_char *) var->data) = (u_char) dat->s[0];
			break;
		case VAR_TYPE_INT:
			*((u_char *) var->data) = (u_char) dat->i[0];
			break;
		case VAR_TYPE_FLOAT:
			*((u_char *) var->data) = (u_char) dat->f[0];
			break;
		case VAR_TYPE_DOUBLE:
			*((u_char *) var->data) = (u_char) dat->d[0];
			break;
		case VAR_TYPE_STRING:
			*((u_char *) var->data) = (u_char) atoi (dat->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a short variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_short (struct jvisu_var *var, union jvisu_intvar *dat)
{

	switch (var->type) {
		case VAR_TYPE_BOOL:
		case VAR_TYPE_BYTE:
			*((short *) var->data) = (short) dat->b[0];
			break;
		case VAR_TYPE_SHORT:
			*((short *) var->data) = dat->s[0];
			break;
		case VAR_TYPE_INT:
			*((short *) var->data) = (short) dat->i[0];
			break;
		case VAR_TYPE_FLOAT:
			*((short *) var->data) = (short) dat->f[0];
			break;
		case VAR_TYPE_DOUBLE:
			*((short *) var->data) = (short) dat->d[0];
			break;
		case VAR_TYPE_STRING:
			*((short *) var->data) = (short) atoi (dat->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a integer variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_int (struct jvisu_var *var, union jvisu_intvar *dat)
{

	switch (var->type) {
		case VAR_TYPE_BOOL:
		case VAR_TYPE_BYTE:
			D1 (printf ("%s var data: %d\n", prompt, (int) dat->b[0]));
			*((int *) var->data) = (int) dat->b[0];
			break;
		case VAR_TYPE_SHORT:
			*((int *) var->data) = (int) dat->s[0];
			break;
		case VAR_TYPE_INT:
			*((int *) var->data) = dat->i[0];
			break;
		case VAR_TYPE_FLOAT:
			*((int *) var->data) = (int) dat->f[0];
			break;
		case VAR_TYPE_DOUBLE:
			*((int *) var->data) = (int) dat->d[0];
			break;
		case VAR_TYPE_STRING:
			*((int *) var->data) = atoi (dat->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a float variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_float (struct jvisu_var *var, union jvisu_intvar *dat)
{

	switch (var->type) {
		case VAR_TYPE_BOOL:
		case VAR_TYPE_BYTE:
			*((float *) var->data) = (float) dat->b[0];
			break;
		case VAR_TYPE_SHORT:
			*((float *) var->data) = (float) dat->s[0];
			break;
		case VAR_TYPE_INT:
			*((float *) var->data) = (float) dat->i[0];
			break;
		case VAR_TYPE_FLOAT:
			*((float *) var->data) = dat->f[0];
			break;
		case VAR_TYPE_DOUBLE:
			*((float *) var->data) = (float) dat->d[0];
			break;
		case VAR_TYPE_STRING:
			*((float *) var->data) = (float) atof (dat->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a double variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_double (struct jvisu_var *var, union jvisu_intvar *dat)
{

	switch (var->type) {
		case VAR_TYPE_BOOL:
		case VAR_TYPE_BYTE:
			*((double *) var->data) = (double) dat->b[0];
			break;
		case VAR_TYPE_SHORT:
			*((double *) var->data) = (double) dat->s[0];
			break;
		case VAR_TYPE_INT:
			*((double *) var->data) = (double) dat->i[0];
			break;
		case VAR_TYPE_FLOAT:
			*((double *) var->data) = (double) dat->f[0];
			break;
		case VAR_TYPE_DOUBLE:
			*((double *) var->data) = dat->d[0];
			break;
		case VAR_TYPE_STRING:
			*((double *) var->data) = atof (dat->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a string variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_string (struct jvisu_var *var, union jvisu_intvar *dat)
{

	switch (var->type) {
	case VAR_TYPE_BOOL:
	case VAR_TYPE_BYTE:
		sprintf ((char *) var->data, "%d", (int) dat->b[0]);
		break;
	case VAR_TYPE_SHORT:
		sprintf ((char *) var->data, "%d", (int) dat->s[0]);
		break;
	case VAR_TYPE_INT:
		sprintf ((char *) var->data, "%d", dat->i[0]);
		break;
	case VAR_TYPE_FLOAT:
		sprintf ((char *) var->data, "%f", dat->f[0]);
		break;
	case VAR_TYPE_DOUBLE:
		sprintf ((char *) var->data, "%f", dat->f[0]);
		break;
	case VAR_TYPE_STRING:
		strcpy ((char *) var->data, dat->b);
		break;
	default:
		return;
	}
}


/** 
 *	Find a matching variable and endianess-convert the incoming data, 
 *	then call the appropriate write function depending on internal
 *	variable type; call the variable specific write function if available
 *
 *	@param dat	receive buffer structure
 *	@param sd	socket descriptor on which the write command was received
 */
static void write_variable (struct jvisu_sndvar	*dat, struct jvisu_sock *sd)
{
	struct list_head 	*lact;
	char			buf[MAX_DATA_LEN];
	union jvisu_intvar 	*idat = (union jvisu_intvar *) buf;
	
	D1 (printf ("%s write variable: idx = %d\n", prompt, je32_to_cpu(dat->idx)));	
	
	/* find the matching entry */	
	list_for_each(lact, &varlist) {
		struct jvisu_var *var = list_entry(lact, struct jvisu_var, list);
		if (var->idx != je32_to_cpu (dat->idx) || var->sd != sd)
			continue;
			
		/* check, if this variable is allowed to be written */
		if ((var->modifytype & VAR_READONLY)) {
			D1 (printf ("%s write variable: request to write readonly var!\n", prompt));
			return;
		}	
			
		D1 (printf ("%s write variable: type: %d, %d\n", prompt, var->type, var->inttype));	
			
		/* do endianess conversion */	
		switch (var->type) {
			case VAR_TYPE_BYTE:
			case VAR_TYPE_BOOL:
				idat->b[0] = dat->data.b[0];
				break;
			case VAR_TYPE_SHORT:
				idat->u16[0] = je16_to_cpu (dat->data.s[0]);
				break;
			case VAR_TYPE_INT:
			case VAR_TYPE_FLOAT:
				idat->u32[0] = je32_to_cpu (dat->data.i[0]);
				break;
			case VAR_TYPE_DOUBLE: {
				idat->u64[0] = je64_to_cpu (dat->data.d[0]);
				break;
			}	
			case VAR_TYPE_STRING:
				dat->data.b[(sd->cmdlen - sizeof (struct jvisu_cmdhdr))] = '\0';
				strcpy (idat->b, dat->data.b);
				break;
			case VAR_TYPE_STREAM:
				break;	
			case VAR_TYPE_BYTEARRAY: 
				break;	
			default:
				return;	
		}
		
		/* call the proper write function*/	
		switch (var->inttype) {
			case VAR_TYPE_BOOL:
				set_bool (var, idat); 
				break;;
			case VAR_TYPE_BYTE:
				set_byte (var, idat); 
				break;
			case VAR_TYPE_SHORT:
				set_short (var, idat); 
				break;
			case VAR_TYPE_INT:
				set_int (var, idat); 
				break;
			case VAR_TYPE_FLOAT:
				set_float (var, idat); 
				break;
			case VAR_TYPE_DOUBLE:
				set_double (var, idat); 
				break;
			case VAR_TYPE_STRING:
				set_string (var, idat); 
				break;
			case VAR_TYPE_STREAM: {
				struct jvisu_varstream *p = (struct jvisu_varstream*) var->data;
				memcpy (p->data, dat->data.b, (sd->cmdlen - sizeof (struct jvisu_cmdhdr)));
				p->len = (sd->cmdlen - sizeof (struct jvisu_cmdhdr));
				break;
				}	
			case VAR_TYPE_BYTEARRAY: {
				/* check length */
				int num = sd->cmdlen - sizeof (struct jvisu_cmdhdr);
				if (num > var->arraylen)
					num = var->arraylen;
				memcpy (var->data, dat->data.b, (size_t) num);
				D1 (printf ("%s receive array: len = %d\n", prompt, num));
				break;
				}
		}
		
		/* check if a write_done function for this variable is available */
		if (var->write_done)
			var->write_done (var->arg);
		
		var->elapsed += var->rate;	
	}
}


/** 
 *	Start socket
 *	
 *	@param port	portnumber to listen
 *	@return		filedescriptor or errorcode
 */
static int start_socket (int port)
{
	int			sd, opt;
	struct sockaddr_in 	sin;	

	/* open the socket */
	sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sd < 0)
		return sd;
		
	/* set non blocking */
	if (fcntl(sd, F_SETFL, O_NONBLOCK) < 0) 
		goto err;

	/* close on exit */
	if (fcntl(sd, F_SETFD, 1) < 0)
		goto err;
	
	/* reuse address */
	opt = 1;
	if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (void *) &opt, sizeof(opt)) < 0)
		goto err;

	/* fill in the neccecary stuff */
	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl (INADDR_ANY);
	sin.sin_port = htons ((uint16_t) port);		

	/* bind it */
	if (bind (sd, (struct sockaddr *) &sin, sizeof (sin)) < 0)
		goto err;
	
	/* try to listen with max connections */
	if (listen(sd, MAX_CONNECTIONS) < 0)
		goto err;

	return sd;
	
err:
	fprintf (stderr, "%s Error in start_socket()\n", prompt);
	close (sd);
	return -1;
}


/** 
 *	Accept connections if we have less than MAX_CONNECTIONS
 *	
 *	@param sd	socket descriptor
 *	@return		new socket descriptor or error
 */
static int do_accept (int sd)
{
	int 			i, fd, opt = 1;
	struct sockaddr_in 	raddr;
	int			addrlen = sizeof (raddr);

	D1 (printf ("%s accepting connection...\n", prompt));

	/* sorry, enough connections */
	if (connections > MAX_CONNECTIONS) {
		D1 (printf ("%s accepting connection: already more than MAX_CONNECTIONS open!\n", prompt));	
		return 0;
	}

	/* try to accept the connection */
  	raddr.sin_family = 0xdead;
	fd = accept(sd, (struct sockaddr *) &raddr, &addrlen);

	if (fd < 0) {
		D1 (printf ("%s accepting connection: failed!\n", prompt));
		if (errno == EAGAIN || errno == EWOULDBLOCK)
			return 0;
		else 	
			return fd;
	}

	/* some linux kernels have this problem when client closes between
	 * select() and accept()	
	 */
	if (raddr.sin_family != AF_INET)
		goto err;

	/* keep alive */
	if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (void *) &opt, sizeof(opt)) < 0)
		goto err;
	/* reuse address */
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *) &opt, sizeof(opt)) < 0)
		goto err;
	/* no delay */
	if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (void *) &opt, sizeof (opt)) < 0)
		goto err;
	
	/* non blocking mode */
	if (fcntl (fd, F_SETFL, O_NONBLOCK) < 0) 
		goto err;

	/* close on exit */
	if (fcntl(fd, F_SETFD, 1) < 0)
		goto err;

	/* put it into an empty place */
	for (i = 0; i <= MAX_CONNECTIONS; i++) {
		if (actsd[i].sd == -1) {
			actsd[i].sd = fd;
			actsd[i].status = 0;
			actsd[i].bufpoi = 0;
			actsd[i].cmdlen = 0;
			connections++;
			D1 (printf ("%s accepting connection: done. socket-fd=%d, currently %d connection(s) active\n", prompt, fd, connections));
			return 0;	
		}
	}
	/* should not happen */
	connections = MAX_CONNECTIONS;
	fprintf (stderr, "%s accepting connection: eeep! no free actsd, but connections < max\n", prompt);
		
	/* error exit for valid socket descriptors */
err:
	fprintf (stderr, "%s accepting connection: error occured, exiting\n", prompt);
	close(fd);
	return 0;
}


/** 
 *	Analyse received command
 *	
 *	@param sd	jvisu socket description structure
 *	@return		0 on success, else error number
 */
static int analyse_command (struct jvisu_sock *sd)
{
	int	res = 0;

	switch (sd->buf[0]) {
	
	/* register variables with server
	 * FIXME: handle string variable
	 */
	case CMD_REG_VAR: {
		struct jvisu_var 	*var;
		struct jvisu_regvar 	*reg = (struct jvisu_regvar *) sd->buf;
		int			namelen;

		/* truncate symbol, if neccecary */
		namelen = MIN (sd->bufpoi - sizeof (struct jvisu_regvar), MAX_SYMBOL_LEN);
		/* terminate string */
		sd->buf[sd->cmdlen] = 0x0;
		var = malloc (sizeof (struct jvisu_var));
		if (!var) {
			res = -ENOMEM;
			break;
		}
		memset (var, 0, sizeof (struct jvisu_var));	
		/* store data */
		var->sd = sd;
		var->type = (int) reg->hdr.type;
		var->idx = je32_to_cpu (reg->idx);
		var->rate = je32_to_cpu (reg->rate);
		var->elapsed = var->rate;
		/* get a pointer to the variable
		 * if it's a bool, get a bitmask too
		 * ignore it, if symbol cannot be retrieved
		 * For byte arrays get the arraysize too
		 */
		if (jops->get_symbol (var, reg->symbol) < 0) {
			free (var);
			res = -EINVAL;
			break;
		}
		
		if (var->inttype == VAR_TYPE_BYTEARRAY)
			var->cmp.p = malloc ((size_t)var->arraylen);
			
		D1 (printf ("%s analyse_command: CMD_REG_VAR(%s), idx=%d\n", prompt, reg->symbol, var->idx));

		list_add_tail (&var->list, &varlist);
		registeredvars++;
		break;
	}
	
	/* FIXME: Handle string variable */
	case CMD_UNREG_VAR: {
		struct list_head *lact, *ltmp;
		struct jvisu_regvar 	*reg = (struct jvisu_regvar *) sd->buf;
		
		D1 (printf("%s analyse_command: CMD_UNREG_VAR(%s), idx = %d\n", prompt, reg->symbol, je32_to_cpu (reg->idx)));
			
		/* remove from list, free memory */
		list_for_each_safe(lact, ltmp, &varlist) {
			struct jvisu_var *var = list_entry(lact, struct jvisu_var, list);
			if (var->idx != je32_to_cpu (reg->idx) || var->sd != sd)
				continue;
			registeredvars--;
			list_del (&var->list);
			if (var->inttype == VAR_TYPE_BYTEARRAY)
				if (var->cmp.p)
					free(var->cmp.p);
			free (var);
			break;	/* break here, as it is unique (hopefully...) */
		}
		break;
	}
	case CMD_CONF_DATA: {
		struct jvisu_sndgen 	*dat = (struct jvisu_sndgen *) sd->buf;

		D1 (printf ("%s analyse_command: CMD_CONF_DATA\n", prompt));
		
		dat->data[(sd->cmdlen - sizeof(struct jvisu_cmdhdr))] = '\0';
		if (jops->save_configs)
			jops->save_configs (dat->hdr.type, dat->data);
		break;
	}
	
	/* write a variable value */
	case CMD_SEND_VAR: {
		struct jvisu_sndvar	*dat = (struct jvisu_sndvar *) sd->buf;
		
		D1 (printf ("%s analyse_command: CMD_SEND_VAR\n", prompt));
		
		write_variable (dat, sd);
		break;
	}
	default:
		break;
	}
	sd->status = 0;
	sd->bufpoi = 0;	
	sd->cmdlen = 0;
	return res;	
}

/** 
 *	Process requests
 *	
 *	@param sd	jvisu socket description structure
 *	@return		0 on success, else error number
 */
static int process_request (struct jvisu_sock *sd)
{
	int	res; 
	size_t	len = sizeof (int);

	/* FIXME: is this correct ? */
	if (sd->status == STAT_WRITE)
		return 0;

	/* check if socket is alive */
	res = read (sd->sd, &sd->buf[sd->bufpoi], 1);
	if (res <= 0) {
		shutdown_sock (sd);
		return 1;
	}
	sd->bufpoi++;
	
	len = sizeof (struct jvisu_cmdhdr);
	
	/* loop until buffer is empty or returns EOF */
	while (1) {	
		if (sd->status == STAT_READ) {
			if (!sd->cmdlen)
				len -= sd->bufpoi;
			else
				len = sd->cmdlen - sd->bufpoi;	
		}
		/* set status to READ */
		sd->status = STAT_READ;	

		/* check for more */
		res = read (sd->sd, &sd->buf[sd->bufpoi], len);
		
		/* handle EOF */
		if (res == EOF)
			return 0;

		/* socket error? */
		if (res < 0) {
			shutdown_sock (sd);
			return 1;
		}

		/* adjust buffer pointer */
		sd->bufpoi += res;
		
		/* have we read the commandlength already ? */
		if (!sd->cmdlen) {
			if (sd->bufpoi < sizeof (struct jvisu_cmdhdr))
				continue;
			sd->cmdlen = (unsigned int) sd->buf[2];
			sd->cmdlen |= ((unsigned int) sd->buf[3]) << 8;
			D1 (printf ("%s processing request: command len=%d bytes\n", prompt, sd->cmdlen));
		}
		
		/* have we got all bytes ? */
		if (sd->bufpoi == sd->cmdlen) {
			D1 (printf ("%s processing request: analysing command %d byte read\n", prompt, sd->cmdlen + sizeof (struct jvisu_cmdhdr)));
			return analyse_command (sd);
		}	
	}
	return 0; /* keep gcc happy */
}

/** 
 * 	Update fdset
 */
static void check_fdset (void)
{
	int	i;
	
	/* check server socket first */
	if (FD_ISSET(actsd[0].sd, &rfds))
		do_accept(actsd[0].sd);
		
	/* server is always set */
	FD_SET(actsd[0].sd, &rfds);

	/* loop through active connections
	 * 0 is always the server socket
	 */
	for (i = 1; i <= MAX_CONNECTIONS; i++) {
		if (actsd[i].sd == -1) 
			continue;
		
		if (actsd[i].status == STAT_WRITE) {
			if (FD_ISSET(actsd[i].sd, &wfds)) {
				actsd[i].status &= ~STAT_WRITE;
			}
			FD_SET (actsd[i].sd, &wfds);
		}
			
		if (FD_ISSET(actsd[i].sd, &rfds)) {
			if (process_request (&actsd[i]))
				continue;
		}
		FD_SET (actsd[i].sd, &rfds);
	}
}


/**
 *	set debug level
 *
 *	@param level	debug level
 *	@return		0 on success
 */
int jvisu_debuglevel (int level)
{	
	jvisu_debug = level;	
	return 0;	
}


/**
 *	start/stop JVisu
 *
 *	@param stop	set to 1 if you want to stop JVisu
 *	@return		0 on success
 */
int jvisu_stop_server (int stop)
{
	jvisu_stop = stop;
	return 0;
}


/** 
 *	JVisu socket server
 *	
 *	@param ops	user supplied functions
 *	@param port	port to listen, if 0 default port is used
 *	@return		0, if a regular stop, errno on error
 */
int jvisu_socket (struct jvisu_ops *ops, int port)
{
	int			res, fd, sd, current = 0;
	struct timeval 		timeo;

	if (!ops || !ops->get_symbol)
		return -EINVAL;

	jops = ops;

	/* initialize variable list */
	INIT_LIST_HEAD (&varlist);
	
	/* open bind and listen on socket */
	sd = start_socket (port ? port : JVISU_PORT);
	if (sd < 0) {
		fprintf (stderr, "%s cannot open socket (port: %d)\n", prompt, port);
		return -EIO;
	}
	D1 (printf ("%s jvisu_socket: active (sd=%d), waiting for connections on port %d..\n", prompt, sd, port));
	
	memset (&actsd, 0xff, sizeof (actsd));	/* preset act sd with -1 */
	actsd[0].sd = sd;			/* store server socket   */ 
	
	/* zero filedescriptor sets */
	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	/* server is always set */
	FD_SET(sd, &rfds);

	/* initialize timekeeping */
	gettimeofday (&tact, NULL);
	
	/* loop until we are stopped */
	while (!jvisu_stop) {
		/* check, if there are active connections */
		fd = 0;
		for (current = 0; current <= MAX_CONNECTIONS; current++) {
			if (actsd[current].sd != -1)
				fd = MAX (actsd[current].sd, fd);
		}
		
		/* if no connections -> come back once a second
		 * to check jvisu_stop; else use the given scan interval 
		 * for variable changes
		 */
		timeo.tv_sec = connections ? 0 : 1;
		timeo.tv_usec = connections ? (JVISU_INTERVAL * 1000) : 0;
		res = select(fd + 1, &rfds, &wfds, NULL, &timeo);

		/* FIXME: can we really continue ? */
		if (res < 0) {
			if (errno == EINTR) continue;
			fprintf (stderr, "%s error (errno=%i) on select: %d\n", prompt, errno, res);
			continue;
		}

		/* check fdsets */
		check_fdset ();				

		/* check variables */
		check_variables ();
	}
	
	D1 (printf ("%s server stopped, shut down sockets\n", prompt));	
	for (current = 0; current <= MAX_CONNECTIONS; current++) {
		if (actsd[current].sd != -1) 
			shutdown_sock (&actsd[current]);
	}		
	return 0;		
}

