package org.jvisu.base;

import java.awt.*;
import java.awt.image.*;
import java.net.*;
import javax.swing.*;
import javax.swing.border.*;
import java.beans.*;
import org.jvisu.panel.*;
import org.jvisu.comm.*;
import org.jvisu.comm.var.*;

/** Visual component to display boolean values
 *
 * $Id: JVBool.java 318 2007-08-21 11:23:19Z mol $
 *  
 *  <p>Copyright 2003 Thomas Gleixner <tglx@linutronix.de></p>
 *
 *  <p>This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.</p>
 * <p>
 *
 * Change Log:
 * </p>
 * <ul>g
 * <li> 1.0 First implementation
 * </ul>
 * <p>
 *  The purpose of JVBool is to display a rectangular shape. 
 *  Depending on a boolean value the appearence of the shape changes. Either 
 *  a given image is displayed or the color of the shape changes.
 *  For both states a text can be assigned.
 *  If the component is not readonly, it acts like a button. 
 *  This can be a toggle button or a simple button depending on the
 *  value of the property isToggle
 * </p>
 * <p>
 * <b>TODO:</b>Implement MNG support for multiframe animation
 * </p>
 * 
 * @author Thomas Gleixner <tglx@linutronix.de>
 * @version 1.0
 * @since JVisu 1.0
 */
public class JVBool extends JVBean implements Runnable {

    /** Image selection: two seperate images for true and false */
    public static final int IMG_TWO_IMAGES = 0;
    /** Image selection: one image (given in imageFalse), which is animated */
    public static final int IMG_ONE_ANIMATED = 1;
    /** Image selection: background (parent) image, which is animated */
    public static final int IMG_BG_ANIMATED = 2;
    
    
    /** Fill color for variable value false. Default is light gray */    
    private Color       falseColor = Color.lightGray;
    /** Fill color for variable value true. Default is green */    
    private Color       trueColor = Color.green;
    /** Text color. Default is black */    
    private Color       textColor = Color.black;
    /** Text assigned to variable value true. Default is null */    
    private String      textTrue = null;
    /** Text assigned to variable value false. Default is null */    
    private String      textFalse = null;
    /** Defines, if the button is a toggle button or a simple button */
    private boolean     isToggle;
    /** Animation time in ms. 0 = off */
    private int         animate;
    /** Animation thread */
    private Thread      animator;
    /** Animation status */
    private boolean     animationState;
    /** unScaled Image for status false */
    private Image       iconFalse;
    /** unScaled Image for status true */
    private Image       iconTrue;
    /** Scaled Image for status false */
    private BufferedImage       scaledImgFalse;
    /** Scaled Image for status true */
    private BufferedImage       scaledImgTrue;
    /** String representing the location of the image file for state false. */
    private String      imageFalse;    
    /** String representing the location of the image file for state true. */
    private String      imageTrue;
    /** Image selector */
    private int         imageSelect;
    /** color for finding the pixels, which have to be modified */
    private Color       animationColor = new Color(0xFFAFAF);
    /** The actual variable value */
    private boolean     actval;
    /** Pixelbuffer */
    private int[]       pixbuf;
    /** Invers boolean */
    private boolean     invers = false;
    /** True is non-transparent */
    private boolean     trueOpaque = true;    
    /** False is non-transparent */
    private boolean     falseOpaque = true;    
    
    
    /** Create a new instance of VisuBool
     */
    public JVBool() {
        addCommVar (new JVVarCfg(null, updaterate, JVVar.T_BOOL, 0, 0));
        addMouseListener(new java.awt.event.MouseAdapter() {
            public void mouseClicked(java.awt.event.MouseEvent evt) {
                if (editEnabled && !readonly && isToggle)
                    setValue (!actval);
            }
            public void mousePressed(java.awt.event.MouseEvent evt) {
                if (evt.getButton() != java.awt.event.MouseEvent.BUTTON1)
                    return;
                if (editEnabled && !readonly && !isToggle)
                    setValue (true);
            }
            public void mouseReleased(java.awt.event.MouseEvent evt) {
                if (evt.getButton() != java.awt.event.MouseEvent.BUTTON1)
                    return;
                if (editEnabled && !readonly && !isToggle)
                    setValue (false);
            }
        });
        this.setDesignsize(new Dimension(20,20));
        this.setOpaque(false);
    }

    /** Startup, start an animation thread, if configured
     */
    public void StartUp() {
        setAnimate (this.animate);
    }

    /** Shutdown, stop the animation, if active
     */
    public void Shutdown() {
        StopAnimator();
    }

    /** Stop the animation thread 
     */
    private void StopAnimator () {
        if (animator != null) {
            animate = 0;
            try { animator.interrupt(); }
            catch (Exception e) {};
            try { animator.join(1000); }
            catch (Exception e) {};
            animator = null;
        }
    }

    /** The animation thread */
    public void run () {
        while (animate > 0) {
            animationState = !animationState;
            repaint();
            try {
                Thread.sleep(animate);
            } catch (InterruptedException e) {}    
        }
    }
    
    /** Set variable value, send it to server, if possible 
    */
    public void setValue (boolean val) {
        setCommVarValue (0, new Boolean (val));
    }
    
    /** Get variable value 
     *  @return the variable content
    */
    public boolean getValue () {
       	return actval;
    }

    /** Property change listener
     *  @param evt the property event
     */
    public void propertyChange(PropertyChangeEvent evt) {
        super.propertyChange(evt);
        String key = evt.getPropertyName();
        if (evt.getSource() == this) {
            if (key.equals("editEnabled") || key.equals("readonly")) {
                if (editEnabled && !readonly)
                    setBorder(new BevelBorder(actval ? BevelBorder.LOWERED : BevelBorder.RAISED));
                else 
                    setFrameColor (getFrameColor());
                repaint();
            }
            if (key.compareTo("base") == 0) {
                readImageTrue();
                readImageFalse();
                repaint();
            }
            /* When we use the parents backgorund image, then react on
             * moving and resizing
             */
            if ((key.equals("preferredSize") || key.equals("designpos")) && imageSelect == 2) {
                scaledImgFalse = null;
                repaint();
            }
        }
    }
        
    /** Informes the owner, that a variable value has been changed
     * @param var Variable whose value has changed
     */
    public void VarChanged(JVVar var) {
        actval = ((JVVarBool) var).getVarValue();
        if (editEnabled && !readonly)
            setBorder(new BevelBorder(actval ? BevelBorder.LOWERED : BevelBorder.RAISED));
        repaint();
    }

    /** Try to read a image from a given icon.
     * If the icon is not defined or not fully loaded, null is returned.
     * If the icon is available and fully loaded a scaled Image, which is retrieved
     * from the ImageIcon is returned.
     * This can be the previous generated Image, if height and width are matching.
     * @param icon String, which shows the place where to read the Image
     * @param img current Image, which is returned, when height and width are matching
     * @param rect rectangle, which contains the display area
     * @return scaled Image fitting into the display are or null, if no Image is available or
     * the Image is not loaded yet.
     */
    private BufferedImage getImage(Image icon, BufferedImage img, Rectangle rect) {
        if (icon == null)
            return null;
        // Have we an existing image ?
        if (img == null || img.getHeight() != rect.height || img.getWidth() != rect.width) {
            img = new BufferedImage (rect.width, rect.height, BufferedImage.TYPE_INT_ARGB);
            Graphics g = img.getGraphics();
            g.drawImage(icon.getScaledInstance(rect.width, rect.height, java.awt.Image.SCALE_FAST), 0, 0, null);
        }
        return img;
    }

    /** Retrieve parent image 
     *  @param rect     rectangle, which identifies the display area
     *  @return scaled image or null, if no image is available
     */
    private BufferedImage getParentImage (BufferedImage img, Rectangle rect) {
        if (img == null || img.getHeight() != rect.height || img.getWidth() != rect.width) {
            try {
                JVPanel panel = (JVPanel) getParent();
                return panel.getComponentImage(this);
            } catch (ClassCastException ec) {
                    return null;
            }    
        }
        return img;
    }
    
    /** Get the image for the selected state
     *  @param state    state for which an image has to be returned
     *  @param rect     rectangle, which identifies the display area
     *  @param color    color for this state
     *  @return scaled image or null, if no image is available
     */
    private BufferedImage getImage (boolean state, Rectangle rect, Color color) {
        BufferedImage img;
        /* Which image handling is selected ? */
        switch (imageSelect) {
            /* Return the image for the given state */
            case IMG_TWO_IMAGES:
                if (state) {
                    scaledImgTrue = getImage(iconTrue, scaledImgTrue, rect);
                    return scaledImgTrue;
                } else {
                    scaledImgFalse = getImage(iconFalse, scaledImgFalse, rect);
                    return scaledImgFalse;
                }
            case IMG_ONE_ANIMATED:
                img = getImage(iconFalse, scaledImgFalse, rect);
                break;
            case IMG_BG_ANIMATED:
                img = getParentImage(scaledImgFalse, rect);
                break;
            default:
                return null;
        }
        scaledImgFalse = img;    
        /* No image available, return */
        if (img == null)
            return null;
        /* If no fillcolor is given, return unmodified image */
        if (color == null || animationColor == null)
            return img;

        /* Check, if the drawing image exists and has the correct size */
        if (scaledImgTrue == null || scaledImgTrue.getHeight() != rect.height || scaledImgTrue.getWidth() != rect.width) {
            scaledImgTrue = new BufferedImage(rect.width, rect.height, java.awt.image.BufferedImage.TYPE_INT_ARGB);
            pixbuf = new int[rect.height * rect.width];
        }
        
        /* Get RGB values */
        int ancol = animationColor.getRGB();
        int fill = color.getRGB();
        /* Read source pixels */
        scaledImgFalse.getRGB(0, 0, rect.width, rect.height, pixbuf, 0, rect.width);
        /* Modify area */
        for (int i = 0; i < pixbuf.length; i++) {
            if (pixbuf[i] == ancol)
                pixbuf[i] = fill;
        }
        /* Write back modified image */
        scaledImgTrue.setRGB(0,0,rect.width, rect.height, pixbuf, 0, rect.width);
        
        return scaledImgTrue;
    }
    
    /** VisuBool specific paint function
     *  Display the state of the variable
     *  If a image is available for this state, display the image
     *  If no image is given, fill the rectangle with the given color
     *  Display text, if available for this state
     * @param g Graphics context
     */
    public void paintComponent (Graphics g) {
        Color col;
        String txt;
        BufferedImage img = null;
        
        Rectangle rect = new Rectangle(getSize());
        // select image, color and text for the given state
        if (!(actval ^ invers)) {
            col = falseOpaque ? falseColor : null;
            txt = textFalse;
            img = getImage(false, rect, col);
        } else {
            txt = textTrue;
            if (animate > 0 && !animationState) {
                col = falseOpaque ? falseColor : null;
                img = getImage(false, rect, col);
            } else {
                col = trueOpaque ? trueColor : null;
                img = getImage(true, rect, col);
            }
        }
        
        // Display image, if available
        if (img != null)
            g.drawImage(img, rect.x, rect.y, null);
        else {
            // Fill component
            if (col != null ) {
                g.setColor (col);
                g.fillRect (rect.x, rect.y, rect.width, rect.height);
            }
        }

        // Display text ?
        if (txt == null) {
            return;
        }
        // set font and textcolor
        g.setColor (textColor);
        g.setFont (getFont());
        
        // retrieve textsize
        int fheight = g.getFontMetrics ().getHeight ();
        int fwidth = g.getFontMetrics ().stringWidth (txt);
        int x = 0; 
        // set x startpoint depending on alignement
        switch (aligncontent) {
            // align left
            case JVAlignment.JVISU_ALIGNLEFT: x = g.getFontMetrics ().charWidth(' '); break;
            // align center
            case JVAlignment.JVISU_ALIGNCENTER: x = (rect.width - fwidth) / 2; break;
            // align right
            default: x = rect.width - fwidth - g.getFontMetrics ().charWidth(' ');
        }
        g.drawString (txt, rect.x + x, rect.y + (rect.height + fheight / 2) / 2);
    }
    
    /** Get whether the fill color for false is opaque or not.
     * @return true if false fill color is oPaque.
     */
    public boolean isFalseOpaque() {
        return falseOpaque;
    }
    
    /** Set whether the the fillcolor for false is opaque or not.
     * @return true if false fill color is oPaque.
     */
    public void setFalseOpaque(boolean opaque) {
        falseOpaque = opaque;
    }

    /** Getter for property falseColor.
     * @return Value of property falseColor.
     */
    public Color getFalseColor() {
        return this.falseColor;
    }
    
    /** Setter for property falseColor.
     * @param falseColor New value of property falseColor.
     */
    public void setFalseColor(Color falseColor) {
        this.falseColor = falseColor;
        repaint();
    }
    
    /** Get whether the fill color for true is opaque or not.
     * @return true if true fill color is oPaque.
     */
    public boolean isTrueOpaque() {
        return trueOpaque;
    }
    
    /** Set whether the the fillcolor for true is opaque or not.
     * @return true if true fill color is oPaque.
     */
    public void setTrueOpaque(boolean opaque) {
        trueOpaque = opaque;
        repaint();
    }

    /** Getter for property trueColor.
     * @return Value of property trueColor.
     */
    public Color getTrueColor() {
        return this.trueColor;
    }
    
    /** Setter for property trueColor.
     * @param trueColor New value of property trueColor.
     */
    public void setTrueColor(Color trueColor) {
        this.trueColor = trueColor;
    }
    
    /** Getter for property textColor.
     * @return Value of property textColor.
     */
    public Color getTextColor() {
        return this.textColor;
    }
    
    /** Setter for property textColor.
     * @param textColor New value of property textColor.
     */
    public void setTextColor(Color textColor) {
        this.textColor = textColor;
    }
    
    /** Getter for property textTrue.
     * @return Value of property textTrue.
     */
    public String getTextTrue() {
        return this.textTrue;
    }
    
    /** Setter for property textTrue.
     * @param textTrue New value of property textTrue.
     */
    public void setTextTrue(String textTrue) {
        this.textTrue = textTrue;
    }
    
    /** Getter for property textFalse.
     * @return Value of property textFalse.
     */
    public String getTextFalse() {
        return this.textFalse;
    }
    
    /** Setter for property textFalse.
     * @param textFalse New value of property textFalse.
     */
    public void setTextFalse(String textFalse) {
        this.textFalse = textFalse;
    }
    
    /** Get the symbol of the communication variable
     * @return the variable symbol of this component
     */
    public String getCommVar () {
        if (invers)
            return "!" + getCommVarSymbol(0);    
        else
            return getCommVarSymbol(0);    
    }

    /** Initialize commvariable
     * @param symbol the variable symbol which is used to retrieve values from the JVisu server
     */
    public void setCommVar (String symbol){
        if (symbol != null && symbol.charAt(0) == '!') {
            invers = true;
            setCommVarSymbol (0, symbol.substring(1));
        } else {
            invers = false;
            setCommVarSymbol (0, symbol);
        }
    }
    
    /** Getter for property isToggle.
     * @return Value of property isToggle.
     */
    public boolean isIsToggle() {
        return this.isToggle;
    }
    
    /** Setter for property isToggle.
     * @param isToggle New value of property isToggle.
     */
    public void setIsToggle(boolean isToggle) {
        this.isToggle = isToggle;
    }
    
    /** Get animation time in ms
     *  @return Value of property imgAnimate.
     */
    public int getAnimate() {
        return this.animate;
    }
    
    /** Set animation time in ms if an animated display should be displayed for status true.
     *  The animation is done by displaying either imgFalse and imgTrue or the fill colors 
     *  alternating with the given animation time delay.<br>
     *  If this value is 0, no animation takes place.<br>
     *  This is useful to implement flashing warnings and alarms or pseudo motion of an image.
     *
     *  @param animate New value of property animate.
     */
    public void setAnimate(int animate) {
        this.animate = animate;
        if (animate < 1)
            StopAnimator();
        else {
            /* The animator thread is started by the Startup method */
            if (!this.isDisplayable())
                return;
            
            if (animator == null || !animator.isAlive()) {
                animator = new Thread (this);
                animator.start();
            }
        }
    }
    
    /** Getter for property imageFalse.
     * @return Value of property imageFalse.
     */
    public String getImageFalse() {
        return this.imageFalse;
    }
    
    /** Setter for property imageFalse.
     * @param imageFalse New value of property imageFalse.
     */
    public void setImageFalse(String imageFalse) {
        this.imageFalse = imageFalse;
        readImageFalse();
    }
    
    /** Read the image file */
    private void readImageFalse() {
        try {
            URL url;
            if (base == null)
                url = new URL (imageFalse);
            else
                url = new URL (new URL(base), imageFalse);
            iconFalse= javax.imageio.ImageIO.read(url);
            scaledImgFalse = null;
        } catch (Exception e) {}
        repaint ();
    }
    /** Getter for property imageTrue.
     * @return Value of property imageTrue.
     */
    public String getImageTrue() {
        return this.imageTrue;
    }
    
    /** Setter for property imageTrue.
     * @param imageTrue New value of property imageTrue.
     */
    public void setImageTrue(String imageTrue) {
        this.imageTrue = imageTrue;
        readImageTrue();
    }

    /** Read the image file */
    private void readImageTrue() {
        try {
            URL url;
            if (base == null)
                url = new URL (imageTrue);
            else
                url = new URL (new URL(base), imageTrue);
            iconTrue= javax.imageio.ImageIO.read(url);
            scaledImgTrue = null;
        } catch (Exception e) {}
        repaint ();
    }

    /** Main method for standalone testing
    public static void main(String args[]) {
        final JFrame frame = new JFrame();
        frame.addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                System.exit(0);
            }
        });
        frame.getContentPane().setLayout(new BoxLayout(frame.getContentPane(), BoxLayout.Y_AXIS));
        final JVBool jb = new JVBool();
        jb.setCommVar("outbit0");
        jb.setServer("localhost");
        jb.setPort(2201);
        jb.setCommtype("org.jvisu.comm.JVSockComm");
        //jb.setReadonly(false);
        //jb.setEditEnabled(true);
        JButton b = new JButton("rmv");
        b.addActionListener(new java.awt.event.ActionListener () {
            public void actionPerformed(java.awt.event.ActionEvent e) {
                frame.getContentPane().remove(jb);
            }
        });
        frame.getContentPane().add(b);
        
        frame.getContentPane().add(jb);
        frame.pack();
        frame.setVisible(true);
    }    
     */
    
    /** Return the actual image selection mode
     * @return image selection mode.
     */
    public int getImageSelect() {
        return this.imageSelect;
    }
    
    /** Set the image selection mode.
     *  0 = use imageFalse and imageTrue
     *  1 = use imageFalse analyse animation color for modifying the content
     *  2 = use parent image, analyse animation color for modifying the content
     *  @param imageSelect New image selection mode.
     */
    public void setImageSelect(int imageSelect) {
        this.imageSelect = imageSelect;
        scaledImgTrue = null;
        scaledImgFalse = null;
        repaint();
    }

    /** Get the color, which is used to identify the modification area inside 
     *  the image.
     *  @return modification area indentification color.
     */
    public Color getAnimationColor() {
        return this.animationColor;
    }
    
    /** Set the color, which is used to identify the modification area inside 
     *  the image.
     *  @param animationColor New modification area identification color.
     */
    public void setAnimationColor(Color animationColor) {
        this.animationColor = animationColor;
    }
}
