package org.jvisu.base;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.net.*;
import java.text.*;
import java.beans.*;
import javax.swing.*;
import org.jvisu.panel.*;
import org.jvisu.comm.*;
import org.jvisu.util.param.JVParams;
import org.jvisu.comm.var.*;

/** Visual component to display a numerical value in an analog gauge
 * <p>
 * $Id: JVGauge.java 318 2007-08-21 11:23:19Z mol $
 * </p>
 *  <p>Copyright 2003 Thomas Gleixner <tglx@linutronix.de></p>
 *
 *  <p>This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation. See file LICENCE in this
 *  directory</p>
 * <p>
 * Change Log:
 * </p>
 * <ul>
 * <li> 1.0 First implementation
 * </ul>
 * <p>
 * </p>
 *  The purpose of JVGauge is to display a value in an analog gauge
 *
 * @author Thomas Gleixner <tglx@linutronix.de>
 * @version 1.0
 */
public class JVGauge extends JVScales {
    
    /** Default inset of the drawing area */
    protected final static int  DEFAULT_INS = 4;
    /** Start angle for this gauge */
    private int                 startAngle = 0;
    /** Apex angle of this gauge */
    private int                 apexAngle = 90;
    /** Text orientation of this gauge */
    private int                 textOrientation = 0;
    /** Text angle of this gauge */
    private int                 textAngle = 0;
    /** Image for the background */
    private Image               bgimg = null;
    /** The rectangle, which contains the arc */
    private Rectangle           arcRect = new Rectangle();
    /** The subimage, which should be read */
    private Rectangle           subRect = new Rectangle();
    /** The offset to draw the subimage */
    private Point               offset = new Point();
    /** The center point on the drawing area */
    private Point               center = new Point();
    /** The needle length */
    private int                 needlelength;
    /** color for the needle */
    private Color               needleColor = Color.blue;
    /** The arc length, which must be filled in % of needle length */
    private int                 arcfilled = 50;
    
    /** Create a new instance of JVLevel
     */
    public JVGauge() {
        this.setDesignsize(new Dimension(200,200));
    }
    
    /** Property change listener
     *  @param evt the property event
     */
    public void propertyChange(PropertyChangeEvent evt) {
        super.propertyChange(evt);
        String key = evt.getPropertyName();
        if (evt.getSource() == this) {
            /* Recalc and repaint */
            if (key.equals("startAngle") ||
                key.equals("apexAngle") ||
                key.equals("textOrientation") ||
                key.equals("textAngle")) {
                recalc = true;
                repaint();
            }
        }
    }
    
    /** JVGauge specific paint function.
     *
     * @param g Graphics context
     */
    public void paintComponent(Graphics g) {
        
        /* Check, if a new buffered background image has to be drawn */
        drawBackgroundImage();
        
        Graphics2D g2 = (Graphics2D) g;
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
        
        /* Get the drawing rectangle */
        Rectangle rect = new Rectangle(getSize());

        if (isOpaque()) {
            g2.setColor(getBackground());
            g2.fillRect(0, 0, rect.width, rect.height);
        }
        
        /* Should we draw a border ? */
        if (drawBorder) {
            border.paintBorder(this, g2, 0, 0, rect.width, rect.height);
        }
        
        /* Calc the needle angle*/
        double start = Math.PI * ((double)((startAngle - 180))) / 180;
        double apex = Math.PI * ((double)apexAngle) / 180;
        
        /* Let the needle be 0.02 % outside of the scale, if the value
         * is out of limits */
        double theta = apex * calcPosition(actval, 0.02);
        
        /* Fill the arc ? */
        if (arcfilled > 0) {
            int fill = needlelength * arcfilled / 100;
            double cmp = (apex < 0) ? -theta : theta;
            g2.setColor(getColorForVal(cmp, Math.abs(apex)));
            int x = center.x - fill;
            int y = center.y - fill;
            g2.fillArc(x, y,  2 * fill, 2 * fill, 180 - startAngle, -(int)Math.toDegrees(theta));
        }
        
        /* Draw the backgorund image */
        g2.drawImage(bgimg, offset.x, offset.y, null);
        
        /* Draw the needle */
        g2.setColor(needleColor);
        g2.transform(AffineTransform.getRotateInstance(start + theta, center.x, center.y));
        g2.drawLine(center.x, center.y, center.x + needlelength, center.y);
    }
    
    /** Draw the background image
     */
    private void drawBackgroundImage() {

        if (!recalc)
            return;
        recalc = false;
        
        /* Normalize angles */
        int beg, end;
        boolean invers = false;
        if (apexAngle > 0) {
            beg = startAngle;
            end = beg + apexAngle;
        } else {
            end = startAngle;
            beg = end + apexAngle;
            invers = true;
        }
        if (beg < 0) {
            beg += 360;
            end += 360;
        }
        
        Rectangle r = new Rectangle(getSize());
        r.grow(-DEFAULT_INS, -DEFAULT_INS);
        
        /* calc the size of the rectangle, which contains the arc */
        /* Begin in first quadrant */
        if (beg >= 0 && beg < 90) {
            if (end > 0 && end <= 90)
                arcRect.setBounds(0, 0, 2, 2);
            if (end > 90 && end <= 180)
                arcRect.setBounds(0, 0, 1, 2);
            if (end > 180)
                arcRect.setBounds(0, 0, 1, 1);
        }
        /* Begin in second quadrant */
        if (beg >= 90 && beg < 180) {
            if (end > 90 && end <= 180)
                arcRect.setBounds(1, 0, 2, 2);
            if (end > 180 && end <= 270)
                arcRect.setBounds(1, 0, 2, 1);
            if (end > 270)
                arcRect.setBounds(0, 0, 1, 1);
        }
        /* Begin in third quadrant */
        if (beg >= 180 && beg < 270) {
            if (end > 180 && end <= 270)
                arcRect.setBounds(1, 1, 2, 2);
            if (end > 270 && end <= 360)
                arcRect.setBounds(0, 1, 1, 2);
            if (end > 360)
                arcRect.setBounds(0, 0, 1, 1);
        }
        /* Begin in forth quadrant */
        if (beg >= 270) {
            if (end > 270 && end <= 360)
                arcRect.setBounds(0, 1, 2, 2);
            if (end > 360 && end <= 450)
                arcRect.setBounds(0, 0, 2, 1);
            if (end > 450)
                arcRect.setBounds(0, 0, 1, 1);
        }
        
        /* Make sure, that the area is a square and fits into the drawing area */
        if (r.width * arcRect.width > r.height * arcRect.height)
            r.width = (r.height * arcRect.height) / arcRect.width;
        else if (r.width * arcRect.width < r.height * arcRect.height)
            r.height = (r.width * arcRect.width) / arcRect.height;
        
        int w = arcRect.width * r.width;
        int shift = 0;
        if (arcRect.width == 2 || arcRect.height == 2)
            shift = (w * 5) / 100;
        w -= 2 * shift;
        
        
        BufferedImage bufimg = new BufferedImage(w, w, BufferedImage.TYPE_INT_ARGB);
        Graphics2D g2 = (Graphics2D) bufimg.getGraphics();
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);

        /* Fill the area transparent */
        g2.setColor(new Color(0, true));
        g2.fillRect(0, 0, w, w);
        
        g2.setColor(getForeground());
        int w2 = w / 2;
        g2.fillOval(w2 - 4, w2 - 4, 8, 8);
        g2.setFont(getFont());
        int scale = g2.getFontMetrics().getHeight();
        
        /* Draw the arc */
        g2.drawArc(scale, scale, w - 2 * scale, w - 2 * scale, -(beg + 180), (beg - end));

        /* Rotate to the start of the scale */
        double t, tact;
        if (!invers)
            t = Math.PI * ((double)(beg + 180)) / 180;
        else
            t = Math.PI * ((double)(end + 180)) / 180;
        g2.transform(AffineTransform.getRotateInstance(t, w2, w2));
        
        t = Math.PI * ((double)(end - beg)) / 180;
        
        int maxFwidth = g2.getFontMetrics().stringWidth(calcMajorText(majortics));
        for (int i = 0; i <= majortics ; i++) {
            if (i != majortics) {
                /* Draw the minor ticks first */
                for (int j = 1; j < minortics; j++) {
                    tact = t * calcMinorTick(i, j, invers);
                    g2.transform(AffineTransform.getRotateInstance(tact , w2, w2));
                    g2.drawLine(w - scale, w2, w - scale/2, w2);
                    g2.transform(AffineTransform.getRotateInstance(-tact , w2, w2));
                }
            }
            /* Draw the major tick */
            tact = t * calcMajorTick(i, invers);
            g2.transform(AffineTransform.getRotateInstance(tact, w2, w2));
            g2.drawLine(w - scale, w2, w - 2, w2);
            /* Draw the text */    
            String txt = calcMajorText(i);
            int fheight = g2.getFontMetrics().getHeight()/4;
            int fwidth = g2.getFontMetrics().stringWidth(txt);


            switch (textOrientation) {
            case 0: /* radial */
                g2.drawString(txt, w - 2 * scale - fwidth, w2 + fheight);
                break;
            case 1: /* radial reverse */
                g2.transform(AffineTransform.getRotateInstance(Math.PI, w2, w2));
                g2.drawString(txt, 2 * scale, w2 + fheight);
                g2.transform(AffineTransform.getRotateInstance(Math.PI, w2, w2));
                break;
            case 2: /* follow */
                g2.transform(AffineTransform.getRotateInstance(Math.PI/2, w - 2 * scale, w2));
                g2.drawString(txt, w - 2 * scale-fwidth/2, w2 + fheight);
                g2.transform(AffineTransform.getRotateInstance(-Math.PI/2, w - 2 * scale, w2));
                break;
            case 3: /* follow reverse*/
                g2.transform(AffineTransform.getRotateInstance(3*Math.PI/2, w - 2 * scale, w2));
                g2.drawString(txt, w - 2 * scale-fwidth/2, w2 + fheight);
                g2.transform(AffineTransform.getRotateInstance(-3*Math.PI/2, w - 2 * scale, w2));
                break;
            case 4:/* angle */
                g2.transform(AffineTransform.getRotateInstance( Math.PI-tact+2*Math.PI*(textAngle-startAngle)/360, w - 2*scale-maxFwidth/2, w2));
                g2.drawString(txt, w - 2*scale-maxFwidth/2 -fwidth/2 , w2 + fheight);
                g2.transform(AffineTransform.getRotateInstance(-Math.PI+tact-2*Math.PI*(textAngle-startAngle)/360, w - 2*scale-maxFwidth/2, w2));
                break;
            }
            /* Go back to origin */
            g2.transform(AffineTransform.getRotateInstance(-tact, w2, w2));
        }

        w = bufimg.getWidth() / arcRect.width;
        int h = bufimg.getHeight() / arcRect.height;
        
        int startx = arcRect.x * w;
        int starty = arcRect.y * h;
        /* Store the center point */
        center.x = (arcRect.width * w) / 2 - arcRect.x * w;
        center.y = (arcRect.height * h) / 2 - arcRect.y * h;

        if (arcRect.width == 2) {
            if (startx != 0)
                startx -= shift;
            w += shift;
            if (arcRect.x != 0)
                center.x += shift;
        }
        if (arcRect.height == 2) {
            if (starty != 0)
                starty -= shift;
            h += shift;
            if (arcRect.y != 0)
                center.y += shift;
        }

        /* Set the bounds and offset into the runtime variables */
        subRect.setBounds(startx, starty, w, h);
        offset.x = (getWidth() - w) / 2;
        offset.y = (getHeight() - h) / 2;
        
        /* Store the center point */
        center.x += offset.x;
        center.y += offset.y;
        /* Store needle length */
        needlelength = w2 - scale / 2;
        
        bgimg = bufimg.getSubimage(subRect.x, subRect.y,subRect.width, subRect.height); 
    }
    
    /** Get the start angle of the gauge
     * @return the start angle
     */
    public int getStartAngle() {
        return startAngle;
    }
    
    /** Set the start angle of the gauge
     * @param start start angle of the gauge
     *
     */
    public void setStartAngle(int start) {
        int old = startAngle;
        startAngle = start;
        firePropertyChange ("startAngle", old, start);
    }
    
    /** Get the apex angle of the gauge
     * @return the apex angle
     *
     */
    public int getApexAngle() {
        return apexAngle;
    }
    
    /** Set the apex angle of the gauge
     * @param apex the apex angle of the gauge
     */
    public void setApexAngle(int apex) {
        int old = apexAngle;
        apexAngle = apex;
        firePropertyChange ("apexAngle", old, apex);
    }
    
    public int getTextOrientation() {
        return textOrientation;
    }
    
    public void setTextOrientation(int orientation) {
        int old = textOrientation;
        textOrientation = orientation;
        firePropertyChange ("textOrientation", old, orientation);
    }
    
    public int getTextAngle() {
        return textAngle;
    }
    
    public void setTextAngle(int angle) {
        int old = textAngle;
        textAngle = angle;
        firePropertyChange ("textAngle", old, angle);
    }
    
    /* The following properties do not affect the 
     * background image. They just affect the drawing of the 
     * current value. So we don't need to repaint 
     * the background image and we fire no propertyChange event
     * A plain repaint is enough.
     */ 
    
    /** Getter for property needleColor.
     * @return Value of property needleColor.
     *
     */
    public Color getNeedleColor() {
        return this.needleColor;
    }
    
    /** Setter for property fillColor.
     * @param fillColor New value of property fillColor.
     *
     */
    public void setNeedleColor(Color color) {
        needleColor = color;
        repaint();
    }

    /** Get the length of the filled arc in percent of the needle length
     *  @return length of the filled arc in percent of the needle length
     */
    public int getArcfilled() {
        return arcfilled;
    }
    
    /** Set the length of the filled arc in percent of the needle length
     *  @param fill length of the filled arc in percent of the needle length
     */
    public void setArcfilled(int fill) {
        arcfilled = fill;
        repaint();
    }
    
    /** Main method for standalone testing */
    public static void main(String args[]) {
        JFrame frame = new JFrame();
        frame.addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                System.exit(0);
            }
        });
        JVGauge lvl = new JVGauge();
        lvl.setCommVar("t1");
        lvl.setServer("localhost");
        lvl.setCommtype("org.jvisu.comm.JVSimRandomComm");
        lvl.setMaxVar("1");
        lvl.setMinVar("0.1");
        lvl.majortics = 9;
        lvl.minortics = 10;
        lvl.setFracDigits(1);
        lvl.startAngle = 180;
        lvl.apexAngle = 90;
        lvl.arcfilled = 50;
        lvl.minalarm = 20;
        lvl.maxalarm = 80;
        lvl.setLogarithmic(true);
        lvl.setFont(new Font("Default", 0, 11));
        lvl.setPreferredSize(new Dimension(300,300));
        lvl.drawBorder = true;
        JPanel p = new JPanel();
        //p.setBackground(Color.gray);
        p.add(lvl);
        frame.getContentPane().add(p);
        frame.pack();
        frame.setVisible(true);
    }
}
