package org.jvisu.comm;

import java.beans.*;
import java.io.*;
import java.net.*;
import java.nio.*;
import java.nio.channels.*;
import java.util.*;
import org.jvisu.comm.var.*;
import org.acplt.ks.*;
import org.acplt.ks.client.*;

/** JVisu ACPLT based communication
 *
 *  <p>$Id: JVAcpltComm.java 169 2003-11-18 08:57:11Z tgl $
 *  </p>
 *  <p>Copyright 2003 Thomas Gleixner <tglx@linutronix.de></p>
 *
 *  <p>This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation. See file LICENCE in this
 *  directory</p>
 *  
 *  <p>Change Log:</p>
 *  <ul>
 *  <li> 1.0 Initial inmplimentation </li>
 *  </ul>
 *  <p>
 *  <p>This class provides the JVisu ACPLT communication functions.
 *  It connects to a ACPLT server and provides the framwork for registering
 *  variables
 *  </p>
 *  <p><b>TODO:</b> Needs more thought. Should use a single thread for
 *  updating.
 *  </p>
 *
 * @author Thomas Gleixner <tglx@linutronix.de>
 * @version 1.0 
 * @since JVisu 1.0
*/
public class JVAcpltComm extends JVComm implements Runnable, PropertyChangeListener {

    /** We share a common package for all instances */
    static protected KscPackage     pkg = null;
    /** Is active */
    private boolean                 isActive;
    /** A thread to manage the update*/
    private Thread                  commThread;

    /** Creates a new instance of JVCommSock
     */
    public JVAcpltComm() {
        JVComm.setStatusMessage(this, 0, "Created");
    }

    /** Property change listener
     *  @param evt the property event
     */
    public void propertyChange(PropertyChangeEvent evt) {
        KscVariable v = (KscVariable) evt.getSource();
        if (v == null)
            return;
        for (int i = 0; i < vars.size(); i++) {
            JVVar var = (JVVar) vars.get(i);
            if (var.getUserObject() != v)
                continue;
            try {
                var.SetVarValue(v.getCurrProps().value.toString());
            } catch (Exception e) { 
                JVComm.setStatusMessage(this, 1, "Error while reading variable:" + var.getSymbol());
                break;
            }
            if (status != 0)
                JVComm.setStatusMessage(this, 0, "OK");
            break;
        }
    }
    
    /** Handle a shared KSC package
     *  @param use true, if the caller wants to start communication
     *             false otherwise
     */
    private static synchronized void handlePackage(boolean use){
        if (use) {
            if (pkg == null) {
                pkg = new KscPackage();
            }
        } else {
           if ((pkg != null) && (pkg.sizeVariables() == 0)) {
                pkg = null;
           }
        }
    }
    
    /** Check, if the instance is matching 
     *  @param cfg configuration to compare
     *  @return true, if matching
     */
    public boolean isMatching (JVCommCfg cfg) {
        return (isActive &&
                cfg.getType() != null &&
                cfg.getType().compareTo(this.getClass().getName()) == 0 &&
                cfg.getServer() != null &&
                cfg.getServer().compareTo(commcfg.getServer()) == 0 &&
                cfg.getPort() == commcfg.getPort());
    }

    /** Check, if the instance is reuseable
     *  @param cfg configuration to compare
     *  @return true, if reuseable
     */
    public boolean isReusable (JVCommCfg cfg) {
        return (!isActive &&
                cfg.getType() != null &&
                cfg.getType().compareTo(this.getClass().getName()) == 0);
    }
    
    /** Check, if the given configuration is valid for
     * this communication type
     * @param cfg the configuration to check
     * @return true, if valid
     */
    protected boolean isValid(JVCommCfg cfg) {
        if (cfg.getServer() == null)
            return false;
        try {
            if (InetAddress.getByName(cfg.getServer()) == null)
                throw new UnknownHostException();
            if (cfg.getPort() < 1024)
                throw new Exception();
        } catch (UnknownHostException e) { 
            JVComm.setStatusMessage(this, 1, "Unknown host");
            return false; 
        } catch (Exception e) {
            JVComm.setStatusMessage(this, 1, "Invalid configuration");
            return false;
        }
        return true;
    }
    
    /** Register a JVisu variable with this instance
     * @param var JVisu variable, which has to be registered
     */    
    protected void RegisterVar(JVVar var) {
        handlePackage(true);
        /* Create an ACPLT/KS variable proxy and connect a listener to
         * the variable's current properties. If the current properties
         * are updated (through the package), then the listener will
         * receive an event and update the display.
         */
        String rl = "//" + commcfg.getServer() + ":" + commcfg.getPort() + "/";
        rl += var.getSymbol();
        try {
            KscVariable v = new KscVariable(rl);
            v.addCurrPropsListener(this);
            pkg.add(v);
            var.setUserObject(v);
        } catch ( KsException e ) {
            JVComm.setStatusMessage(this, 1, "Error when creating a new variable: " + var.getSymbol());
        }
    }
    
    /** Unregister a JVisu variable with this instance
     * @param var JVisu variable, which has to be unregistered
     */    
    protected void UnRegisterVar(JVVar var) {
        KscVariable v = (KscVariable) var.getUserObject();
        if (v != null)
            pkg.remove(v);
        var.setUserObject(null);
    }

    /** Send a variable value to the server
     * @param var JVisu variable which has to be sent to the server
     * @return true on success
     */    
    public boolean SendVar(JVVar var) {
        if (!isConnected())
            return false;

        return true;    
    }
    
    /** Get the activation status
     * @return true, if active
     */    
    public boolean isActive() {
        return isActive; 
    }
    
    /** Run the connection thread
     */
    public void run() {
        while (isActive) {
            DoCommunication ();
            // if we come back, wait 1 second before trying to reconnect
            try { Thread.sleep (1000); }
            catch (InterruptedException e) {break;}
        }
    }    

    /** Start communication */
    protected void startComm() {
        if (commThread == null) {
            commThread = new Thread(this);
            isActive = true;
            commThread.start();
            JVComm.setStatusMessage(this, 0, "Started");        
        }
    }
    
    /** Stop communication to server.
     *  This can only be done, if all variables are unregistered.
     */    
    protected boolean stopComm() {
        isActive = false;
        setConnected (false);
        try { commThread.interrupt(); }
        catch (Exception e) {};
        try { commThread.join(1000); }
        catch (Exception e) {};
        commThread = null;
        handlePackage(false);
        JVComm.setStatusMessage(this, 0, "Stopped");        
        return true;
    }
    
    /** Do the communication stuff
     */
    private void DoCommunication () {
        try {
            synchronized (instsem) {
                if (pkg != null)
                    pkg.getUpdate();
                setConnected (true);
                //status = "OK";
            }
        } catch ( KsException e ) { 
            JVComm.setStatusMessage(this, 1, "Error on package update");
        }
    }
}
