package org.jvisu.comm.var;

import java.lang.Object;
import org.jvisu.comm.*;

/** Configuration storage for JVisu Variable.
 *
 *  <p>package:  org.jvisu.comm.var</p>
 *  <p>file:    JVVarCfg.java</p>
 *  <p>$Id: JVVarCfg.java 302 2005-10-16 10:43:06Z ero $ </p>
 *  <p>Copyright 2003 Thomas Gleixner <tglx@linutronix.de></p>
 *
 *  <p>This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation. See file LICENCE in this
 *  directory</p>
 *  
 *  <p>Change Log:</p>
 *  <ul>
 *  <li> 1.0 Initial inmplimentation </li>
 *  </ul>
 *  <p>
 *  Store all variable related configuation information
 *  </p>
 *
 * @author Thomas Gleixner <tglx@linutronix.de>
 * @version 1.0 
*/
public class JVVarCfg extends Object{
    
    /** Variable symbol */
    private String      varSymbol;
    /** Variabletype corresponding to the variable address */
    private int         varType;
    /** Updaterate for this variable */
    private int         varRate;
    /** Size information for this variable (Array types only) */
    private int         varSize;
    /** private data depending on variable type */
    private int         varPrivate;
    /** declares, if this variable does an forced update in the given time */
    private boolean     varForcedUpdate = false;
    /** declares, if this variable is enabled or not, true by default */
    private boolean     varEnabled = true;
    /** The variable object, which is associated to this configuration */
    private JVVar       varInst;
    
    /** Create a new JVVarCfg
     * @param symbol        a string representing of the Variable symbol in the variable server
     * @param updaterate    the requested max. updaterate for this variable in ms
     * @param type          the variable type, which is requested from the server
     * @param size          the variable size for array types
     * @param priv          type dependend private information
     */
    public JVVarCfg(String symbol, int updaterate, int type, int size, int priv) {
        varSymbol   = symbol;
        varRate     = updaterate;
        varType     = type;
        varSize     = size;
        varPrivate  = priv;
    }

    /** Create a new JVVarCfg and copy the contents of the old one.
     *  Everything is copied, except the communication instance reference
     *  @param cfg Configuration data to clone
     */
    public JVVarCfg(JVVarCfg cfg) throws NullPointerException {
        varSymbol   = cfg.getSymbol();
        varRate     = cfg.getRate();
        varType     = cfg.getType();
        varSize     = cfg.getSize();
        varPrivate  = cfg.getPrivate();
        varEnabled  = cfg.isEnabled();
    }
    
    /** Checks, if the variable configuration is matching 
     *  @param cfg a variable configuration, which should be compared
     *  @return true if matching
     */
    public boolean isMatching (JVVarCfg cfg) {
        return (cfg.getSymbol() != null && 
                cfg.getSymbol().equals(varSymbol) &&
                cfg.getType() == varType && 
                cfg.isForcedUpdated() == varForcedUpdate
                );
    }
    
    /** Returns the variable symbol
     *  @return String which contains symbol or null
     */
    public String getSymbol() {
        return varSymbol;
    }

    /** Set the variable symbol
     *  @param sym String which contains symbol or null
     */
    public void setSymbol(String sym) {
        varSymbol = new String(sym);
    }
    
    /** Returns the variable type
     *  @return variable type
     */
    public int getType() {
        return varType;
    }
    
    /** Returns the variable updaterate
     *  @return updaterate
     */
    public int getRate() {
        return varRate;
    }

    /** Set the variable updaterate
     *  @param rate updaterate
     */
    public void setRate(int rate) {
        varRate = rate;
    }

    /** Returns the variable array size
     *  @return variable array size
     */
    public int getSize() {
        return varSize;
    }

    /** Returns the variable array size
     *  @return variable array size
     */
    public void setSize(int size) {
        varSize = size;
    }

    /** Returns the variable private data
     *  @return private data
     */
    public int getPrivate() {
        return varPrivate;
    }

    /** Return the variable enabled state
     *  @return true, if variable is enabled
     */
    public boolean isEnabled() {
        return varEnabled;
    }

    /** Set the variable enabled state
     *  @param ena true, if variable is enabled
     */
    public void setEnabled(boolean ena) {
        varEnabled = ena;
    }

    /** Return wether teh variable must be updated
     *  in the given updaterate, regardless if the variable 
     *  has changed or not
     *  @return true, if variable has a forced update mode
     */
    public boolean isForcedUpdated() {
        return varForcedUpdate;
    }

    /** Set wether the variable must be updated
     *  in the given updaterate, regardless if the variable 
     *  has changed or not
     *  @param force true, if variable has forced update mode
     */
    public void setForcedUpdated(boolean force) {
        varForcedUpdate = force;
    }

    /** Get the variable instance, which is associated to this configuration
     *  @return variable instance
     */
    public JVVar getVarInstance() {
        return varInst;
    }

    /** Set the variable instance, which is associated to this configuration
     *  @param inst variable instance
     */
    public void setVarInstance(JVVar inst) {
        varInst = inst;
    }
}
