package org.jvisu.panel.configurator;

import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.*;
import java.beans.*;
import javax.swing.*;
import javax.swing.border.*;
import org.jvisu.util.*;
import org.jvisu.util.param.*;
import org.jvisu.panel.configurator.events.*;

/** Panel property 
 *  <p>package:    org.jvisu.panel</p>
 *  <p>file:       JVPropertyCellPanel.java</p>
 * $Id: JVPropertyCellPanel.java 318 2007-08-21 11:23:19Z mol $
 *
 *  <p>Copyright 2003 Thomas Gleixner <tglx@linutronix.de></p>
 *
 * <p>This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.</p>
 * </p>
 *
 * Change Log:
 * </p>
 * <ul>
 * <li> 1.0 First implementation
 * </ul>
 * <p>
 * </p>
 * JVPropertyCellPanel is a panel to display the contents of a single property
 * and provide the editor for this property
 *
 * @author Thomas Gleixner <tglx@linutronix.de>
 * @version 1.3
 */
public class JVPropertyCellPanel extends JVCellPanel implements ActionListener, PropertyChangeListener {
    
    /** Component, which has to be introspected */
    private JComponent          editComponent;
    /** Property */
    private JVPalettePanel.PaletteItemProperty  propdescr;
    /** Property */
    private PropertyDescriptor  property;
    /** Property editor */
    private PropertyEditor      propEdit;
    /** edit Class  */
    private Class               editClass;
    /** property Class  */
    private Class               propClass;
    /** inner Rectangle */
    private Rectangle           paintarea;
    /** Component, which is used to edit the property */
    private JComponent          editor;
    /** Object, representing this property */
    private Object              propObj;
    /** Button to start custom editor */
    private JButton             startEditor;
    /** The property editor, which provides custom dialogs */
    JVPanelPropertyEditor       owner;
    
    /** Creates a new instance of JVPanelPropertyEditor 
     *  @param comp component, which is edited
     *  @param prop property descriptor of this component
     *  @param size size of this component
     *  @param font font descriptor
     */
    public JVPropertyCellPanel(JComponent comp, JVPalettePanel.PaletteItemProperty prop, Dimension size, Font font,
            JVPanelPropertyEditor owner) {
        super(size, font);
        paintarea = new Rectangle();
        this.owner = owner;
        setProperty (comp, prop);
    }

    /** Called by the parent component, when the component is
     *  removed.
     *  Shut down the component and free resources
     */
    public void removeNotify() {
        /* Let the super know what happened */
        super.removeNotify();
        if (editComponent == null || property == null)
            return;
        editComponent.removePropertyChangeListener(property.getName(), this);
        removeAll();
        editComponent = null;
        editor = null;
    }
    
    /** Sets the properties of this panel
     *  @param comp component, which is edited
     *  @param prop property descriptor of this component
     */
    public void setProperty(JComponent comp, JVPalettePanel.PaletteItemProperty prop) {
        editComponent = comp;
        propdescr = prop;
        
        try {
            property = new PropertyDescriptor(prop.getName(), comp.getClass());
        } catch(IntrospectionException e) {
            return;
        }
        
        if (property.getReadMethod() == null || property.getWriteMethod() == null)
            return;
        
        /* Get the Object of this property */
        try {
            propObj = property.getReadMethod().invoke(comp, (Object[])null);
        } catch (Exception e) { return; }
    
        /* Keep us informed, when the property changes */
        comp.addPropertyChangeListener(property.getName(), this);
        
        /* Property class */
        propClass = property.getPropertyType();
        
        /* Find a property editor for this class */
        try { 
            editClass = Class.forName(prop.getEditor());
        } catch (Exception e) {
            editClass = property.getPropertyEditorClass();
        }

        if (editClass != null) {
            try {
                propEdit = (PropertyEditor) editClass.newInstance();
            } catch (Exception e) {
                try {
                    Class args[] = { Object.class };
                    Constructor cons = editClass.getConstructor(args);
                    Object obj[] = { propObj };
                    propEdit = (PropertyEditor)cons.newInstance(obj);
                } catch (Exception e1) {
                    propEdit = null;
                }
            }
        }
        
        if (propEdit == null)
            propEdit =  PropertyEditorManager.findEditor(propClass);
        
        if (propEdit != null) { 
            /* Set the objectb value */
            propEdit.setValue(propObj);
            
            if (propEdit.getTags() != null) {
                JComboBox box = new JComboBox(propEdit.getTags());
                box.setFont(getFont());
                editor = box;
                box.addActionListener(new ActionListener() {
                    public void actionPerformed (java.awt.event.ActionEvent evt) {
                        try {
                            propEdit.setAsText((String)((JComboBox)editor).getSelectedItem());
                            Object[] args = new Object[] { propEdit.getValue() };
                            property.getWriteMethod().invoke(editComponent, args);        
                        } catch (Exception e) { }
                        updateProperty();
                    }
                });
                add(editor, BorderLayout.CENTER);
                
            } else if (owner.hasCustomDialog (propClass)) {
            
                startEditor = new JButton("...");
                add(startEditor, BorderLayout.EAST);
                startEditor.addActionListener(this);

            } else if (!propEdit.isPaintable()) {
                JTextField edit = new JTextField();
                
                if (propObj == null)
                    edit.setText("");
                else
                    edit.setText(propEdit.getAsText());
                
                edit.setFont(getFont());
                editor = edit;
                add(edit, BorderLayout.CENTER);
                edit.addActionListener(new ActionListener() {
                        public void actionPerformed (java.awt.event.ActionEvent evt) {
                            try {
                                propEdit.setAsText((String)((JTextField)editor).getText());
                                Object[] args = new Object[] { propEdit.getValue() };
                                property.getWriteMethod().invoke(editComponent, args);        
                            } catch (Exception e) { 
                                System.out.println(e);
                            }
                            updateProperty();
                        }
                });
            }
            updateProperty();
        }
    }
    
    /** Return the property value as String 
     *  @param comp component of same type to compare with default value
     *  @return property value as String or null, if it is the default value
     */
    public String getProperty (JComponent comp) { 
        if (editComponent == null)
            return null;
        try {
            propObj = property.getReadMethod().invoke(editComponent, (Object[])null);
            Object deflt = property.getReadMethod().invoke(comp, (Object[])null);
            
            if (propClass.isAssignableFrom(Font.class)) {
                Font fn = (Font) propObj;
                Font fd = (Font) deflt;
                if (fn.getName().compareTo(fd.getName()) == 0 &&
                        fn.getSize() == fd.getSize() &&
                        fn.getStyle() == fd.getStyle())
                    return null;
                return (fn.getName() + "," + fn.getSize() + "," + fn.getStyle());
            }
            
            if (propObj.equals(deflt))
                return null;
            if (propEdit.getAsText() != null)
                return propEdit.getAsText();
            
            return propObj.toString();
        } catch (Exception e) { 
            return null; 
        }
    }
    
    /** Update property value
     */
    private void updateProperty() {
        if (editComponent == null)
            return;
        /* Get the Object of this property */
        try {
            propObj = property.getReadMethod().invoke(editComponent, (Object[])null);
            propEdit.setValue(propObj);
        } catch (Exception e) { return; }

        if (editor instanceof JComboBox) {
            ((JComboBox)editor).setSelectedItem(propEdit.getAsText());
            
        } else if (owner.hasCustomDialog (propClass)) {
            
        } else if (!propEdit.isPaintable()) {
            ((JTextField)editor).setText(propEdit.getAsText());
        }
        editComponent.repaint();
        repaint();
    }
    
    /** Action listener */
    public void actionPerformed (java.awt.event.ActionEvent evt) {
        
        if (evt.getSource() == startEditor) {
            if (owner.hasCustomDialog (propClass)) {
                Object obj = owner.showCustomDialog(propClass,propEdit.getValue());
                if (obj != null) {
                    try {
                        propEdit.setValue(obj);
                        Object[] args = new Object[] { propEdit.getValue() };
                        property.getWriteMethod().invoke(editComponent, args);        
                    } catch (Exception e) { 
                        System.out.println(e);
                    }
                }
                updateProperty();
            }
        }
    }

    /** Property change listener
     *  @param evt the property event
     */
    public void propertyChange(PropertyChangeEvent evt) {
        
        if (evt.getSource() == editComponent) {
            String key = evt.getPropertyName();
            if (key.equals(property.getName())) {
                updateProperty();
                //System.out.println (evt.getOldValue() + " , " + evt.getNewValue());
                String oldval = JVParams.ObjectToString(evt.getOldValue());
                if (oldval != null)
                    owner.fireEditEvent(new JVEditEvent(owner, JVEditEvent.PROP_CHANGED, 
                        editComponent, property.getName(), oldval));
            }
        }
    }
    
    /** Paint component */
    public void paintComponent(Graphics g) {
        /* If we have no Object given, let super paint an empty panel */
        if (propObj == null || editor != null) {
            super.paintComponent(g);
            return;
        }
            
        /* Do we have a property editor ?? */
        if (propEdit != null) {
            if (propEdit.isPaintable()) {
                Insets ins = getInsets();
                paintarea.x = ins.left;
                paintarea.y = ins.top;
                paintarea.width = this.getWidth() - ins.right;
                paintarea.height = this.getHeight() - ins.bottom;
                propEdit.paintValue(g, paintarea);
            }
        }
    }
}

